/*
Copyright Jan Panteltje 2003

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define MAX_SCREEN_LINES	4	// limitation OSD memory?

/* picture size */
#define MAX_H	720
#define MAX_V	576

/* osd devices */
#define MAX_DEV	10

/* OSD windows */
#define LOGO_WINDOW			1
#define SUBTITLE_WINDOW		2
#define MAX_WINDOWS			8	// see TMS 7110 data sheet 


#include "osdplay.h"

FILE *filefd;

extern int debug_flag;
static long long old_stop_milliseconds;
static int subtitle_osd_open_flag;


char *strsave(char *s) /* save string s somewhere */
{
char *p;

p = malloc(strlen(s) + 1);
if(!p) return 0;
strcpy(p, s);
return p;
} /* end function strsave */


unsigned char *img;
int xd, yd, x0, y0;

#ifndef O_BINARY 
#define O_BINARY 0 //for mingw
#endif

typedef struct
{
unsigned char r, g, b, t;
} palt;

palt pal_[4], pal[4];
palt epal[4];

typedef struct
	{
	unsigned short int BM;
	unsigned int size;
	unsigned int r0;
	unsigned int ofs;
	unsigned int hsize;
	unsigned int x;
	unsigned int y;
	unsigned short int bplanes;
	unsigned short int bpp;
	unsigned int comp;
	unsigned int imgsize;
	unsigned int xpm,ypm;
	unsigned int cused;
	unsigned int cimp; 
	} __attribute__ ((packed)) bmpt;


int read_bmp(char *name)
{
bmpt bmp;
int fdr;
unsigned int a, d, i, x, y, size, pals, cst[256];
unsigned short int b;
unsigned char c;
unsigned char *limg;
palt lpal[256];
int pallette_start;

for(i = 0; i < MAX_H * MAX_V; i++)
	{
	img[i] = 0;
	}

fdr = open(name, O_RDONLY | O_BINARY);
if(fdr < 0)
	{
	fprintf(stderr,"error, unable to open file %s\n",name);
	return -1;
	} 
 
if(read(fdr, &bmp, 54) < 54)
	{
	fprintf(stderr,"error, file too short: %s\n",name);
	close(fdr);
	return -1;
	}	
 
if(bmp.BM  !=  19778)
	{
	fprintf(stderr,"error, file %s isn't a bmp\n",name);
	close(fdr);
	return -1;
	}
 
if(debug_flag) fprintf(stderr,"xd: %d  yd: %d\n", bmp.x, bmp.y);
 
xd = bmp.x;
yd = bmp.y;

if( (bmp.x > 720) || (bmp.y > 576) )
	{
	fprintf(stderr,"error, image larger than 720x576\n");
	close(fdr);
	return -1;
	} 

if( (bmp.bpp  !=  4)  &&  (bmp.bpp  !=  8) )
	{
	fprintf(stderr,"error, only 4 & 8bpp images supported\n");
	close(fdr);
	return -1;
	}

if(bmp.comp)
	{
	fprintf(stderr,"error, only uncompressed images are supported\n");
	close(fdr);
	return -1;
	}

d = 0;
 
//lseek(fdr, bmp.hsize - 40, SEEK_CUR);
 
/*
start color table
file header size + info header size
*/

pallette_start = 14 + bmp.hsize; // hsize = BitmapInfoHeader size!

if(debug_flag)
	{
	fprintf(stderr, "read_bmp(): name=%s pallette_start=%d\n", name, pallette_start);
	}

lseek(fdr, pallette_start, SEEK_SET);

pals = 1  <<  bmp.bpp;

for(b = 0; b < pals; b++)
	{
	d += read(fdr, &lpal[b].b, 1); 
	d += read(fdr, &lpal[b].g, 1); 
	d += read(fdr, &lpal[b].r, 1); 
	d += read(fdr, &c, 1);
	if(debug_flag)
		{
		fprintf(stderr, "read_bmp(): lpal[b].b=%d lpal[b].g=%d lpal[b].r=%d c=%d d=%d\n",\
		lpal[b].b, lpal[b].g, lpal[b].r, c, d);
		}
	}
 
if(d != pals * 4)
	{
	fprintf(stderr,"error file too short: %s\n",name);
	close(fdr);
	return -1;
	}	
 
if(bmp.bpp  ==  4)
	{
	if(xd & 7) size = (xd & (-8) ) + 8;
	else size = xd;
  
	size *= yd;
	size /= 2;

	limg = malloc(size);
  
	lseek(fdr, bmp.ofs, SEEK_SET);
  
	if( read(fdr, limg, size)  !=  size)
		{
		fprintf(stderr,"error, file too short: %s\n",name);
		free(limg);
		close(fdr);
		return -1;
		}	
 
	d = 0;
	for(y = yd - 1; y  !=  -1; y--)
		{
		for(x = 0; x < xd; x += 2)
			{
			img[y * xd + x] = limg[d]  >>  4;
			if(x + 1  ==  xd)
				{
				/* should clear img[...] here? */
				d++;
				}
			else img[y * xd + x + 1] = limg[d++] & 15;
			}
		if(d & 3) d += 4 - (d & 3);
		} 
	} /* end if 4 bpp */
else /* 8 bits per pixel */
	{
 	if(xd & 3) size = (xd & (-4) ) + 4;
	else size = xd;
  
	size *= yd;

	limg = malloc(size);
  
	lseek(fdr, bmp.ofs, SEEK_SET);
  
	if(read(fdr, limg, size)  !=  size)
		{
		fprintf(stderr,"error, file too short: %s\n",name);
		free (limg);
		close(fdr);
		return -1;
		}	
	d = 0;
	for(y = yd - 1; y  !=  -1; y--)
		{
		for(x = 0; x < xd; x++) img[y * xd + x] = limg[d++];

		if(d & 3) d += 4 - (d & 3);
		} 
	} /* end if 8 bpp */
 
free(limg);
close(fdr);
 
//#define P_TEST
#ifdef P_TEST

for(a = 0; a < 4; a++)
	{
	pal[a].r = lpal[a].r;
	pal[a].g = lpal[a].g;
	pal[a].b = lpal[a].b;
	pal[a].t = lpal[a].t;

//	if(debug > 3)
		{
		fprintf(stderr, "read_bmp(): P_TEST setting\n\
		pal[a].r=%d\n\
		pal[a].g=%d\n\
		pal[a].b=%d\n\
		lpal[a].t=%d\n",\
		pal[a].r,\
		pal[a].g,\
		pal[a].b,\
		lpal[a].t);
		}
	}

a = 0;
return 0;
#endif

if(debug_flag)
	{
	for(i = 0; i < pals; i++)
		{
		fprintf(stderr,\
		"read_bmp(): lpal[%d].b=%d lpal[%d].g=%d lpal[%d].r=%d c=%d d=%d\n",\
		i, lpal[i].b, i, lpal[i].g, i, lpal[i].r, c, d);
		}
	}

for(a = 0; a < pals; a++) cst[a] = 0;

a = 0;
for(y = 0; y < yd; y++)
	{
	for(x = 0; x < xd; x++)
		{
/*
		fprintf(stderr, "read_bmp(): a=%d img[a]=%d cst[%d]=%d\n",\
		a, img[a], img[a], cst[img[a]]);
*/

		cst[ img[a++] ]++;
		}
	}

if(debug_flag)
	{
	fprintf(stderr, "read_bmp(): name=%s pals=%d yd=%d xd=%d\n", name, pals, yd, xd);
	}

b = 0;
for(a = 0; a < pals; a++)
	{
	if(debug_flag)
		{
		fprintf(stderr, "read_bmp(): cst[%d]=%d b=%d\n", a, cst[a], b);
		}

	if(cst[a])
		{
		if(b > 3)
			{
			if(debug_flag) fprintf(stderr,\
"warning: more than 4 colors found in bitmap %s, setting color %d to 0\n",\
name, a);
			lpal[a].t = 0;
			}
		else
			{
		   	pal[b].r = lpal[a].r;
			pal[b].g = lpal[a].g;
			pal[b].b = lpal[a].b;

			if(debug_flag)
				{
				fprintf(stderr, "read_bmp(): 3 setting\n\
				b=%d a=%d\n\
				pal[b].r=%d lpal[a].r=%d\n\
				pal[b].g=%d lpal[a].g=%d\n\
				pal[b].b=%d lpal[a].b=%d\n",\
				b, a,\
				pal[b].r, lpal[a].r,\
				pal[b].g, lpal[a].g,\
				pal[b].b, lpal[a].b);
				}

			lpal[a].t = b++;

			if(debug_flag)
				{
				fprintf(stderr, "read_bmp(): 3a setting\n\
				lpal[a].t=%d\n",
				lpal[a].t);
				}
			} /* end if b <= 3 */
		} /* end if cst[a] */
	} /* end for all pallettes */ 

a = 0;
for(y = 0; y < yd; y++)
	{
	for(x = 0; x < xd; x++)
		{
		img[a++] = lpal[ img[a] ].t;
//		img[a++] = pal[ img[a] ].t;
		}
	}

return 0;
} /* end function read_bmp */


int title_init()
{
int a, i;
char temp[1024];

if(debug_flag)
	{
	fprintf(stdout, "title_init(): arg none\n");
	}

if(osd_device_name)
	{
	osd_channel = open(osd_device_name, O_RDWR);
	if(osd_channel == -1)
		{
		fprintf(stdout, "Could not open %s for write, aborting\n", osd_device_name);

		exit(1);
		}

	if(debug_flag)
		{
		fprintf(stdout, "Using %s\n", osd_device_name);
		}

	}
else/* default osd devices */
	{
	/* find a free osd device */
	for(i = 0; i < MAX_DEV; i++)
		{
		sprintf(temp, "/dev/ost/osd%d", i);
		osd_channel = open(temp, O_RDWR);

		if(osd_channel != -1)
			{
//			if(debug_flag)
				{
				fprintf(stdout, "Using %s\n", temp);
				}
	
			break;
			}

		fprintf(stdout, "Could not open %s for write\n", temp);
		} /* end while find free video device */

	/* abort if not found */
	if(osd_channel == -1)
		{
		fprintf(stdout, "No free osd device found, aborting\n");

		exit(1);
		}
	}

if(debug_flag)
	{
	fprintf(stdout, "title_init(): using osd_channel=%d\n", osd_channel);
	}

clear_titles(osd_channel);

if(logo_flag)
	{
	logo(osd_channel);

	if(! selected_project) return 1;
	}

a = OSDSetWindow(osd_channel, SUBTITLE_WINDOW);
if(debug_flag)
	{
	fprintf(stdout, "add_graphics(): OSDSetWindow=%d\n", a);
	}

img = malloc(MAX_H * MAX_V);
if(! img)
	{
	fprintf(stdout, "title_init(): could not allocate space for img\n");
	}

subtitle_osd_open_flag = 0;

/* open the .sub file */
sprintf(temp, "%s/.subtitles/%s/%s.sub", home_dir, selected_project, selected_project);
subfileptr = fopen(temp, "r");
if(! subfileptr)
	{
	fprintf(stdout, "Could not open file %s for read, anborting\n", temp);

	return 0;
	}

old_stop_milliseconds = LONG_MAX;

/* get start time */
gettimeofday(start_timeval, NULL);
while(1)
	{	
	get_next_subtitle();

    /* take it easy, more apps are runing */
    usleep(10000);
	}

return 1;
} /* end function title_init */


int get_next_subtitle()
{
int a, b, c, i, j;
char temp[4096];
char bitmapfile[4096];
int start_hour, start_minute, start_second, start_hundreds;
int stop_hour, stop_minute, stop_second, stop_hundreds;
int hsize, vsize;
int hstart, vstart;
int transp0, transp1, transp2, transp3;
long long start_milliseconds, stop_milliseconds;
long long start, now, elapsed;
int osd_clear_flag;
int eof_flag;
double da;
int inc;
int bits_per_pixel;

if(debug_flag)
	{	
	fprintf(stdout, "get_next_subtitle(): arg none\n");
	}

/* 
filename      start       end         hsize vsize hstart vstart transp transp transp transp
frame3944.bmp 00:02:37:76 00:02:39:68 432   48    149    514     0      15     15    0
*/

if(vob2sub_flag)
	{
	a = fscanf(subfileptr, "%s  %d:%d:%d,%d %d:%d:%d,%d  %d %d  %d %d  %d %d %d %d",\
	bitmapfile,\
	&start_hour, &start_minute, &start_second, &start_hundreds,\
	&stop_hour, &stop_minute, &stop_second, &stop_hundreds,\
	&hsize, &vsize,\
	&hstart, &vstart,\
	&transp0, &transp1, &transp2, &transp3); 
	}
else /* default xste .sub file */
	{
	a = fscanf(subfileptr, "%s  %d:%d:%d:%d %d:%d:%d:%d  %d %d  %d %d  %d %d %d %d",\
	bitmapfile,\
	&start_hour, &start_minute, &start_second, &start_hundreds,\
	&stop_hour, &stop_minute, &stop_second, &stop_hundreds,\
	&hsize, &vsize,\
	&hstart, &vstart,\
	&transp0, &transp1, &transp2, &transp3); 
	}

if(a == EOF)
	{
	fprintf(stdout, "get_next_subtitle(): EOF found in subfile, stop\n");
	eof_flag = 1;
	}
else eof_flag = 0;

if(! eof_flag)
	{
	if(! vob2sub_flag && a != 17) 
		{
		fprintf(stdout,\
		"get_next_subtitle(): parse error in subfile line %d, only % d of 13 arguments read, aborting\n",\
		subfile_line, a);

		exit(1);
		}

	subfile_line++;

	/*
	fprintf(stdout, "start_hour=%d start_minute=%d start_second=%d start_hundreds=%d\n",
	start_hour, start_minute, start_second, start_hundreds);
	fprintf(stdout, "stop_hour=%d stop_minute=%d stop_second=%d stop_hundreds=%d\n",
	stop_hour, stop_minute, stop_second, stop_hundreds);
	*/

	//if(debug_flag)
		{
		fprintf(stdout,\
		"get_next_subtitle(): line=%05d %s %02d:%02d:%02d:%02d %02d:%02d:%02d:%02d  %03d %03d  %03d %03d  %02d %02d %02d %02d\n",\
		subfile_line,\
		bitmapfile,\
		start_hour, start_minute, start_second, start_hundreds,\
		stop_hour, stop_minute, stop_second, stop_hundreds,\
		hsize, vsize,\
		hstart, vstart,\
		transp0, transp1, transp2, transp3);
		}

	sprintf(temp, "%s/.subtitles/%s/%s", home_dir, selected_project, bitmapfile);
	if( read_bmp(temp) == -1)
		{
		fprintf(stdout, "could not read bitmap, aborting\n");

		exit(1);
		}

	if(debug_flag)     
		{
		fprintf(stderr, "get_next_subtitle(): after read_bmp(): xd=%d yd=%d x0=%d y0=%d\n", xd, yd, x0, y0);
		}

	hsize = xd;
	vsize = yd;

	if(debug_flag)
		{
		fprintf(stdout, "get_next_subtitle(): hsize=%d vsize=%d\n", hsize, vsize);
		}

	start_milliseconds =\
		(start_hour * 3600000) + (start_minute * 60000) + (start_second * 1000) + (start_hundreds * 10); 
	stop_milliseconds =\
		(stop_hour * 3600000) + (stop_minute * 60000) + (stop_second * 1000) + (stop_hundreds * 10); 

	/* add command line specified offset */
	start_milliseconds += offset_milliseconds;
	stop_milliseconds += offset_milliseconds;

	if( (start_milliseconds  < 0) || (stop_milliseconds < 0) )
		{
		fprintf(stdout, "offset causes negative time, aborting\n");
		fprintf(stdout, "offset_milliseconds=%ld\n", offset_milliseconds);
		fprintf(stdout, "start_milliseconds=%ld\n", start_milliseconds);
		fprintf(stdout, "stop_milliseconds=%ld\n", stop_milliseconds);

		exit(1);
		}

	if(debug_flag)
		{
		fprintf(stdout, "pal[0].r=%d pal[0].g=%d pal[0].b=%d\n", pal[0].r, pal[0].g, pal[0].b);
		fprintf(stdout, "pal[1].r=%d pal[1].g=%d pal[1].b=%d\n", pal[1].r, pal[1].g, pal[1].b);
		fprintf(stdout, "pal[2].r=%d pal[2].g=%d pal[2].b=%d\n", pal[2].r, pal[2].g, pal[2].b);
		fprintf(stdout, "pal[3].r=%d pal[3].g=%d pal[3].b=%d\n", pal[3].r, pal[3].g, pal[3].b);
		}

	if(bw_flag) /* force black and white subtitles */
		{
		OSDSetColor(osd_channel, 0,   0,   0,   0, 0);   // background, black, transparent
		OSDSetColor(osd_channel, 1, 240, 240, 240, 255); // body, white
		OSDSetColor(osd_channel, 2,   0,   0,   0, 255); // outline1, black
		OSDSetColor(osd_channel, 3,   0,   0,   0, 255); // outline2, black, normally not used
		}
	else /* color */
		{
		OSDSetColor(osd_channel, 0, pal[0].r, pal[0].g, pal[0].b, transp0 * 16); 
		OSDSetColor(osd_channel, 1, pal[1].r, pal[1].g, pal[1].b, transp1 * 16);
		OSDSetColor(osd_channel, 2, pal[2].r, pal[2].g, pal[2].b, transp2 * 16);
		OSDSetColor(osd_channel, 3, pal[3].r, pal[3].g, pal[3].b, transp3 * 16);
		}
	} /* end if ! eof_flag */

/* test if OSC clear needed */
if(old_stop_milliseconds < start_milliseconds)
	{
	osd_clear_flag = 1;
	}
else osd_clear_flag = 0;

/* time in milli seconds */
start = (start_timeval -> tv_sec * 1000) + (start_timeval -> tv_usec / 1000);
while(1)
	{
	gettimeofday(current_timeval, NULL);

	now = (current_timeval -> tv_sec * 1000) + (current_timeval -> tv_usec / 1000);

	elapsed = now - start;

	if(debug_flag)
		{
		fprintf(stdout, "elapsed=%ld\n", elapsed);
		fprintf(stdout, "speed_factor=%f\n", speed_factor);
		}

	da = (double) elapsed * (double)speed_factor; 
	elapsed = (long long) da;

	if(debug_flag)
		{
		fprintf(stdout, "corrected elapsed=%ld\n", elapsed);
		}

	if(debug_flag)
		{
		fprintf(stdout, "start=%ld\n", start);
		fprintf(stdout, "now=%ld\n", now);
		fprintf(stdout, "elapsed=%ld\n", elapsed);
		fprintf(stdout, "start_milliseconds=%ld\n", start_milliseconds);
		fprintf(stdout, "stop_milliseconds=%ld\n", stop_milliseconds);
		fprintf(stdout, "old_stop_milliseconds=%ld\n", old_stop_milliseconds);
		fprintf(stdout, "osd_clear_flag=%d\n", osd_clear_flag);
		}

	if(no_sync_flag) break;

	/* display part */
	if(osd_clear_flag || eof_flag)
		{
		/* wait for time to clear */
		if(elapsed >= old_stop_milliseconds)
			{
			if(debug_flag)
				{
				fprintf(stdout, "get_next_subtitle() stop\n");
				}

			/* close window */
			OSDClose(osd_channel);
			subtitle_osd_open_flag = 0;
			osd_clear_flag = 0;
			}
		}
	else
		{
		/* wait for time to display subtitle */
		if(elapsed >= start_milliseconds) break;
		}

	/* 10 milli seconds to reduce processor cycles */
	usleep(10000);
	}

/* exit OK if EOF */
if(eof_flag) exit(0);

/* clear current subtitle */
//OSDClear(osd_channel);

/* close current subtitle area */
if(subtitle_osd_open_flag) OSDClose(osd_channel);

bits_per_pixel = 8;
/* display new subtitle */
a = OSDOpen(osd_channel, hstart, vstart, hstart + hsize, vstart + vsize, bits_per_pixel, 0);
if(a == -1)
	{
	fprintf(stdout, "add_graphics(): could not open OSD, aborting\n");

	exit(1);
	}
subtitle_osd_open_flag = 1;

if(debug_flag)
	{
	fprintf(stdout,\
	"add_graphics(): hstart=%d vstart=%d bits_per_pixel=%d vsize=%d hsize=%d img=%p\n",
	hstart, vstart, bits_per_pixel, hsize, vsize, img);
	}

usleep(10000);

inc = hsize;
a = OSDSetBlock(osd_channel, 0, 0, hsize, vsize - 2, inc, img);
if(debug_flag)
	{
	fprintf(stdout, "add_graphics(): OSDSetBlock=%d\n", a);
	}

old_stop_milliseconds = stop_milliseconds;

return 1;
} /* end function get_next_subtitle */


int clear_titles(int channel)
{
int a, i;
int text_windows;

if(debug_flag)
	{
	fprintf(stdout, "clear_titles(): arg channel;=%d\n", channel);
	}

text_windows = 3;

/* clear all text windows (but not the logo window) */
for(i = 0; i < text_windows; i++)
	{
	a = OSDSetWindow(channel, i);
	if(debug_flag)
		{
		fprintf(stdout, "clear_titles(): i=%d OSDSetWindow=%d\n", i, a);
		}

	/* close window */
	OSDClose(channel);
	} /* end for i (all text windows) */

return 1;
} /* end function clear_titles */


int logo(int channel)
{
int a, b, i, j;
int x, y;
int xsize, ysize;
static unsigned char bitmap[MAX_H * MAX_V];
unsigned char *ptr;
int pos;

/* ADD YOUR OWN LOGO HERE WITH THE bitmap PROGRAM */

#define logo6_width 16
#define logo6_height 16
static unsigned char logo6_bits[] =
{
0xff, 0x01, 0xff, 0x07, 0x03, 0x06, 0x03, 0x06, 0x03, 0x06, 0x03, 0x06,
0xff, 0x07, 0xff, 0x01, 0x03, 0x00, 0x03, 0x00, 0xc3, 0x03, 0x03, 0x02,
0x03, 0x02, 0x03, 0x02, 0x43, 0x02, 0x83, 0x01
};

if(debug_flag)
	{
	fprintf(stdout, "logo(): arg none, using channel=%d\n", channel);
	}

/* THESE NEED TO BE SET FOR YOUR BITMAP */
x = 650;				// x positionin pixels
y = 50;					// y position pixels
xsize = logo6_width;	// x size of bitmap (normally 16)		
ysize = logo6_height;	// y size of bitmap (normally 16)
ptr = logo6_bits;		// pointer to bitmap

a = OSDSetWindow(channel, LOGO_WINDOW);
if(debug_flag) 
	fprintf(stdout, "logo(): OSDSetWindow=%d\n", a);

a = OSDOpen(channel, x, y, x + xsize, y + ysize, 32, 0);
if(a == -1)
	{
	fprintf(stdout, "logo(): could not open OSD, aborting\n");

	exit(1);
	}

/*
note the first 4 colors are used by the subtitles,
make your own logo colors here in red, green, blue values from 0-255
This is just an example.
You have 4 colors available.
*/
//                      r  g  b  transparency 
OSDSetColor(channel, 4, 0, 0, 0, 255);			// black
OSDSetColor(channel, 5, 240, 240, 240, 255);	// almost white
OSDSetColor(channel, 6, 240, 0, 0, 255);		// red
OSDSetColor(channel, 7, 0, 240, 0, 255);		// green

pos = 0;
for(i = 0; i < xsize * ysize / 8; i++)
	{
	/* 8 bits of the byte */
	for(j = 0; j < 8; j++)
		{
		a = *ptr & (1 << j);
	
		if(a) bitmap[pos + j] = 5;
		else bitmap[pos + j] = 6;

		pos ++;
		}
	ptr++;
	}

a = OSDSetBlock(channel, 0, 0, xsize, ysize, xsize, bitmap);
if(debug_flag)
	fprintf(stdout, "logo(): OSDSetBlock=%d\n", a);

return 1;
} /*end function logo */


